/*
 * xdls.c:
 * List keys in an xdata database.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdls.c,v 1.7 2001/03/25 02:34:50 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <sys/types.h>

#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>

#include "util.h"
#include "xdata.h"

void usage() {
    fprintf(stderr,
"xdls: list keys in an xdata database\n"
"\n"
"usage: xdls [-h] [-p path] [-t type] [-i] [-l] [key | /regexp/] ...\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -i        Names and regular expressions will match case-insensitively.\n"
"\n"
"   -l        Show sizes and modification times for keys.\n"
"\n"
"   -e        Show internal and metadata keys instead of user keys.\n"
"\n"
"The given keys, or keys matching the given regexps, if present, will be\n"
"listed on standard output.\n"
"\n"
"If any key is listed, xdls will exit succesfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

const char optstring[] = "+hp:t:ile";

struct match {
    char *str;
    regex_t re;
};

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH");
    char *key;
    int c;
    XDATA xd;
    int ret = 1;
    int nmm, i;
    struct match *mm;
    int nocase = 0, showmore = 0, showinternal = 0;

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xdls: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xdls: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;

            case 'i':
                nocase = 1;
                break;

            case 'l':
                showmore = 1;
                break;

            case 'e':
                showinternal = 1;
                break;
                
            default:
                fprintf(stderr, "xdls: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xdls: no xdata path specified\n");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xdls: no xdata type specified\n");
        usage();
    }
    
    /* Get the things to match. */
    nmm = argc - optind;

    if (nmm > 0) {
        int j;
        mm = (struct match*)calloc(nmm, sizeof(struct match));
        for (i = 0, j = optind; i < nmm; ++i, ++j) {
            if (*argv[j] == '/' && *(argv[j] + strlen(argv[j]) - 1) == '/') {
                /* Argument is a regular expression. */
                char *x;
                x = strdup(argv[j] + 1);
                *(x + strlen(x) - 1) = 0;
                if ((i = regcomp(&mm[i].re, x, REG_EXTENDED | REG_NOSUB | (nocase ? REG_ICASE : 0)))) {
                    char errbuf[1024];
                    regerror(i, &mm[i].re, errbuf, sizeof(errbuf));
                    fprintf(stderr, "xdls: %s: %s\n", argv[j], errbuf);
                    regfree(&mm[i].re);
                    exit(1);
                }
                free(x);
            } else mm[i].str = argv[j];
        }
    }

    /* Start xdata. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xdls: cannot open xdata\n");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 0, NULL)) {
        fprintf(stderr, "xdls: %s\n", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }

    /* Walk through the database, listing keys. */
    key = NULL;
    while (xdata_enum(xd, showinternal, &key)) {
        int f = !nmm;
        for (i = 0; i < nmm; ++i) {
            if ((mm[i].str && (((nocase ? strcasecmp(mm[i].str, key): strcmp(mm[i].str, key)) == 0)))
                || (!mm[i].str && regexec(&mm[i].re, key, 0, NULL, 0) == 0))
                f = 1;
        }
        if (f) {
            /* Match. */
            if (showmore) {
                time_t mtime;
                size_t len;
                if (!xdata_stat(xd, key, &len, &mtime))
                    fprintf(stderr, "xdls: key `%s': %s\n", key, xd->errmsg);
                else {
                    char timestr[128];
                    struct tm *t;
                    t = localtime(&mtime);
                    if (abs(mtime - time(NULL)) > (365 * 24 * 3600))
                        strftime(timestr, sizeof(timestr), "%b %e  %Y", t);
                    else
                        strftime(timestr, sizeof(timestr), "%b %e %H:%M", t);

                    printf("%6d %s %s\n", (int)len, timestr, key);
                }
            } else printf("%s\n", key);
            ret = 0;
        }
    }
    
    xdata_close(xd);
    exit(ret);
}
