/*
 * xdput.c:
 * Put some data into an xdata store.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdput.c,v 1.6 2001/03/25 23:31:32 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>

#include "xdata.h"


const char optstring[] = "+hp:t:rc";

void usage() {
    fprintf(stderr,
"xdput: put data into an xdata database\n"
"\n"
"usage: xdput [-h] [-p path] [-t type] [-r] [-c] key [file]\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -r        Specifies that existing data of the given key will be\n"
"             replaced, if present.\n"
"\n"
"   -c        Specifies that a new database will be created, if none exists.\n"
"\n"
"The contents of file, or data read from stdin if it is not specified, will\n"
"be stored in the database for type under the name key.\n"
"\n"
"If the data are stored, xdput will exit successfullly.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH"), *key, *file;
    int replace = 0, create = 0;
    int c;
    char *data, *p;
    size_t buf_len = 0, len;
    int fd;
    XDATA xd;
    ssize_t l;
    int ret = 1;

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xdput: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xdput: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;
                
            case  'r':
                replace = 1;
                break;

            case 'c':
                create = 1;
                break;

            default:
                fprintf(stderr, "xdput: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xdput: no xdata path specified\n");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xdput: no xdata type specified\n");
        usage();
    }
    
    key = argv[optind++];
    if (!key) {
        fprintf(stderr, "xdput: no key specified\n");
        usage();
    }

    if (strncmp(key, "$$:", 3) == 0) {
        fprintf(stderr, "xdput: key `%s': keys beginning `$$:' are reserved\n", key);
        exit(1);
    }
    
    file = argv[optind];
    if (file) {
        fd = open(file, O_RDONLY);
        if (fd == -1) {
            fprintf(stderr, "xdput: %s: %s\n", file, strerror(errno));
            exit(1);
        }
    } else {
        file = "standard input";
        fd = 0;
    }

    /* OK, try to get hold of data. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xdput: cannot open xdata\n");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 1, create ? xd->path[0] : NULL)) {
        fprintf(stderr, "xdput: %s\n", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }

    /* If we are not in replace mode, we need to check for the existence of
     * the data.
     */
    if (!replace && xdata_exists(xd, key)) {
        if (xd->status)
            fprintf(stderr, "xdput: key `%s': %s\n", key, xd->errmsg);
        else
            fprintf(stderr, "xdput: key `%s': data exists\n", key);
        xdata_close(xd);
        exit(1);
    }

    /* Read data from fd. */
    buf_len = 1024;
    len = 0;
    p = data = (char*)malloc(buf_len);
    do {
        l = read(fd, p, buf_len - (data - p));
        if (l == -1 && errno != EINTR) {
            /* XXX what about EPIPE? */
            fprintf(stderr, "xdput: %s: %s\n", file, strerror(errno));
            xdata_close(xd);
            exit(1);
        }
        if (l > 0) {
            len += l;
            p += l;
            if (p == data + buf_len) {
                l = p - data;
                data = realloc(data, buf_len * 2);
                buf_len *= 2;
                p = data + l;
            }
        }
    } while (l);

    /* Now insert it into the database. */
    if (!xdata_put(xd, key, data, len, time(NULL), replace))
        fprintf(stderr, "xdput: key `%s': %s\n", key, xd->errmsg);
    else ret = 0;

    xdata_close(xd);
    exit(ret);
}
